f_AUTO_RELOAD_DEBUG = true 

local NUMBER_OF_SYNTHS = 2
local NUMBER_OF_SOURCES_PER_SYNTH = 8

local array_instrument_synth = {}

local array_buffer_sources_for_each_synth = {}
local array_buffer_destination = {}
  
local array_sample_sources_for_each_synth = {}
local array_sample_destination = {}
	
local array_bool_bad_setup = {}

local array_parameters_for_each_synth = {}

local array_bool_mixing_requested = {}

function init()

  print("init")
  
  array_instrument_synth = {}

  array_buffer_sources_for_each_synth = {}
  array_buffer_destination = {}

  array_sample_sources_for_each_synth = {}
  array_sample_destination = {}
  
  array_parameters_for_each_synth = {}
  array_bool_bad_setup = {}  
  
  if array_parameters_for_each_synth then
  
    for int_synth = 1, NUMBER_OF_SYNTHS do
	
	  if array_parameters_for_each_synth[int_synth] then
	  
	    for int_parameter = 1, table.getn(array_parameters_for_each_synth[int_synth]) do
		
		  local parameter = array_parameters_for_each_synth[int_synth][int_parametetr]
		  
		  if parameter.value_observable:has_notifier(parameter_changed, array_parameter_info) then
		  
		    parameter.value_observable:remove_notifier(parameter_changed, array_parameter_info)
		  
		  end
		
		end
		
	  end
	  
	end
       
  end

  for int_instrument = 1, table.getn(renoise.song().instruments) do
    local instrument = renoise.song().instruments[int_instrument];

	if(instrument.samples_observable:has_notifier(look_for_synths_list_change)) then
	  instrument.samples_observable:remove_notifier(look_for_synths_list_change)
    end
	  
	if not instrument.name_observable:has_notifier(look_for_synths_instrument_renaming) then
	  
	  instrument.name_observable:add_notifier(look_for_synths_instrument_renaming)
	  
	  for int_sample = 1, table.getn(instrument.samples) do
	    local sample = instrument.samples[int_sample]
		
		if sample.name_observable:has_notifier(look_for_synths_source_renaming) then
		  sample.name_observable:remove_notifier(look_for_synths_source_renaming)
		end
		if sample.name_observable:has_notifier(look_for_synths_destination_renaming) then
		  sample.name_observable:remove_notifier(look_for_synths_destination_renaming)
		end
		
	  end
	
	end
	
  end

  for int_synth = 1, NUMBER_OF_SYNTHS do
  
    array_buffer_sources_for_each_synth[int_synth] = {}
    array_sample_sources_for_each_synth[int_synth] = {}
    array_parameters_for_each_synth[int_synth] = {}
    array_bool_mixing_requested[int_synth] = false
	
  end  
  
end

function get_source(int_synth, int_source, int_frame)

  --print("get_source" .. int_synth .. " " .. int_source .. " " .. int_frame)

  local buffer_source = array_buffer_sources_for_each_synth[int_synth][int_source]
  local buffer_destination = array_buffer_destination[int_synth]
  
  if 
    not buffer_source or 
	not buffer_source.has_sample_data or
    not buffer_destination or 
	not buffer_destination.has_sample_data
  then 
    print("Something went wrong with sample data")
    return nil
  end
	
  local int_frames_source = buffer_source.number_of_frames
  local int_frames_destination = buffer_destination.number_of_frames
  
  local int_rate_source = buffer_source.sample_rate
  local int_rate_destination = buffer_destination.sample_rate
  
  local int_channels_source = buffer_source.number_of_channels
  local int_channels_destination = buffer_destination.number_of_channels
  
  --local real_frame_ratio = int_frames_source * int_rate_source / (int_frames_destination * int_rate_destination)
  local real_frame_ratio = int_frames_source / (int_frames_destination)
  
  --print("Frame ratio for source #" .. int_source .. ": " .. real_frame_ratio)
  
  local int_frame_source = math.max(math.min(int_frame * real_frame_ratio, int_frames_source),1)

  local array_real_frame_channels = {}
  
  if int_channels_source == int_channels_destination then
  
    --print("Source #" .. int_source .. " and destination have both " .. int_channels_source .. " channels")
	  
    for int_channel = 1, int_channels_source do		
      array_real_frame_channels[int_channel] = buffer_source:sample_data(int_channel,int_frame_source)
	  --print(int_source,int_channel, int_frame, int_frame_source, array_real_frame_channels[int_channel])
    end

  elseif int_channels_source == 1 then -- Source is mono, Destination is stereo

    --print("Source #" .. int_source .. " has " .. int_channels_source .. " channels; Destination has " .. int_channels_destination)  
	
	local real_value_source = buffer_source:sample_data(int_channels_source,int_frame_source)
	
	-- spread source into a stereo source
	for int_channel = 1, int_channels_destination do		
	  array_real_frame_channels[int_channel] = real_value_source
	end
		
  elseif int_channels_source == 2 then -- Source is stereo, Destination is mono

    --print("Source #" .. int_source .. " has " .. int_channels_source .. " channels; Destination has " .. int_channels_destination)  
  	local real_value_source = 0
  
	-- mix source into a mono source
	for int_channel = 1, int_channels_source do		
	  real_value_source = real_value_source + buffer_source:sample_data(int_channel,int_frame_source)
	end

	array_real_frame_channels[1] = real_value_source / int_channels_source
	
  end
  
  return array_real_frame_channels
  
end

function mix(int_synth)

  --print("Mix Synth #"..int_synth)
  local instrument = array_instrument_synth[int_synth]
  local sample_destination = array_sample_destination[int_synth]
  local buffer_destination = array_buffer_destination[int_synth]
  local buffer_source1 = array_buffer_sources_for_each_synth[int_synth][1]
  
  if (not sample_destination) or (not buffer_source1) then
    return nil
  end
  
  if  
    not buffer_destination or
    not buffer_destination.has_sample_data or 
	buffer_destination.number_of_channels ~= buffer_source1.number_of_channels or
	buffer_destination.sample_rate ~= buffer_source1.sample_rate or
	buffer_destination.number_of_frames ~= buffer_source1.number_of_frames
  then
  
    if instrument.name_observable:has_notifier(look_for_synths_instrument_renaming) then
	  instrument.name_observable:remove_notifier(look_for_synths_instrument_renaming)
	end
	for int_sample = 1, table.getn(instrument.samples) do
	  local sample = instrument.samples[int_sample]
	  if sample.name_observable:has_notifier(look_for_synths_source_renaming) then
	    sample.name_observable:remove_notifier(look_for_synths_source_renaming)
	  end
	end
    if sample_destination.name_observable:has_notifier(look_for_synths_destination_renaming) then
	  sample_destination.name_observable:remove_notifier(look_for_synths_destination_renaming)
	end
    if instrument.samples_observable:has_notifier(look_for_synths_list_change) then
	  instrument.samples_observable:remove_notifier(look_for_synths_list_change)
	end
  
    print("Had to create sample data for synth#" .. int_synth .. ", sample " .. sample_destination.name)
	buffer_destination:create_sample_data(
	  buffer_source1.sample_rate, 
	  buffer_source1.bit_depth, 
	  buffer_source1.number_of_channels, 
	  buffer_source1.number_of_frames
	) 
    print("Sample data for synth#" .. int_synth .. " created")
	
	instrument.name = "Synth"
	sample_destination.name = "Synth destination"	
	sample_destination.name_observable:add_notifier(look_for_synths_destination_renaming)
	if not instrument.samples_observable:has_notifier(look_for_synths_list_change) then
	  instrument.samples_observable:add_notifier(look_for_synths_list_change)
	end
	if not instrument.name_observable:has_notifier(look_for_synths_instrument_renaming) then
	  instrument.name_observable:add_notifier(look_for_synths_instrument_renaming)
	end
	for int_source = 1, table.getn(array_buffer_sources_for_each_synth[int_synth]) do
	  local sample_source = array_sample_sources_for_each_synth[int_synth][int_source]
	  if not sample_source.name_observable:has_notifier(look_for_synths_source_renaming) then
	    sample_source.name_observable:add_notifier(look_for_synths_source_renaming)
	  end
	end

  end
  
  buffer_destination:prepare_sample_data_changes()

  local int_sources = table.getn(array_buffer_sources_for_each_synth[int_synth])
  local int_valid_sources = 0
  local real_parameter_normalization = 0
  local array_bool_valid_parameter = {}
  
  for int_source = 1, int_sources do

    local real_parameter = array_parameters_for_each_synth[int_synth][int_source]
	--print(real_parameter)
    if real_parameter and real_parameter > 0 then

      real_parameter_normalization = real_parameter_normalization + real_parameter
      int_valid_sources = int_valid_sources + 1
	  array_bool_valid_parameter[int_source] = true

	else
	
	  array_bool_valid_parameter[int_source] = false
	
    end
  
  end
  
  if int_valid_sources > 0 then
    int_valid_sources = 1 / int_valid_sources
    real_parameter_normalization = 1 / (real_parameter_normalization * int_valid_sources)
  else
    real_parameter_normalization = 0
  end
	
  --print(real_parameter_normalization, int_valid_sources)
  
  local array_real_normalized_parameters = {}
  for int_parameter, real_value in pairs(array_parameters_for_each_synth[int_synth]) do
    array_real_normalized_parameters[int_parameter] = real_value * real_parameter_normalization
  end

  for int_frame = 1, buffer_destination.number_of_frames do

    local real_frame = {0,0}
  
    if not (real_parameter_normalization == 0 or int_valid_sources == 0) then
	
      for int_source = 1, int_sources do
	
	    if array_bool_valid_parameter[int_source] then
	  
          local real_frame_source = get_source(int_synth, int_source, int_frame)
		
		  if real_frame_source then
		    for int_channel = 1, table.getn(real_frame_source) do
		      real_frame[int_channel] = 
		        real_frame[int_channel] + 
			     real_frame_source[int_channel] * 
			       array_real_normalized_parameters[int_source] 
		    end
		  end

	    end
	  
      end
	  
	end
	
	for int_channel = 1, buffer_destination.number_of_channels do
	
	  --print(int_channel,int_frame,real_frame[int_channel])
	
	  buffer_destination:set_sample_data(
	    1, 
		int_frame, 
		real_frame[int_channel] * int_valid_sources
	  )
	  
	end

  end
	
  buffer_destination:finalize_sample_data_changes()
  
end

function parameter_changed(array_parameter_info)

  local parameter = array_parameter_info[1]
  local int_synth = array_parameter_info[2]
  local int_source = array_parameter_info[3]

  if array_parameters_for_each_synth and array_parameters_for_each_synth[int_synth] then
    array_parameters_for_each_synth[int_synth][int_source] = parameter.value / 127
  end
  if array_bool_mixing_requested then
    array_bool_mixing_requested[int_synth] = true
  end

end

function look_for_synths_list_change()
  print("look_for_synths_list_change")
  look_for_synths()
end
function look_for_synths_instrument_renaming()
  print("look_for_synths_instrument_renaming")
  look_for_synths()
end
function look_for_synths_source_renaming()
  print("look_for_synths_source_renaming")
  look_for_synths()
end
function look_for_synths_destination_renaming()
  print("look_for_synths_destination_renaming")
  look_for_synths()
end

function look_for_synths()

  init()

  local int_cont = 0
  local int_instruments = table.getn(renoise.song().instruments)
  
  local int_synths = 0
  array_bool_bad_setup[int_synths] = true

  for int_instrument = 1, int_instruments do 
  
    local instrument = renoise.song().instruments[int_instrument]
	
    if not instrument.name_observable:has_notifier(look_for_synths_instrument_renaming) then
      instrument.name_observable:add_notifier(look_for_synths_instrument_renaming)
	  --print(int_cont .. " hadn't got it, now it has got it")
	else
  	  --print(int_cont .. " already has got it")
    end
	
	if instrument.name == "Synth" and int_synths < NUMBER_OF_SYNTHS - 1 then
	
	  int_synths = int_synths + 1
	
	  array_instrument_synth[int_synths] = instrument
	
	  local int_samples = table.getn(instrument.samples)
	  
	  if int_samples < 1 then
	  
	    print("not enough samples in the instrument '" .. instrument.name .. "'")
	    bool_bad_setup[int_synths] = true -- setup is incomplete and nothing can ben done
		
  	  else 

  	    local int_sources = 1
		
		local sample_destination = nil
		local buffer_source = nil
			  
	    for int_sample = 1, int_samples do
		
	      local sample = instrument.samples[int_sample]
		  
		  if sample then
		  
		    print(sample.name)
			
		    if 
		      not array_sample_destination[int_synths] and 
		      sample.name == "Synth destination" 
		    then
			
  		      print("Got destination for synth #" .. int_synths)
		      array_sample_destination[int_synths] = sample
		      array_buffer_destination[int_synths] = sample.sample_buffer
			  
		    elseif 
			  sample.name == "Synth source" and 
			  sample.sample_buffer.has_sample_data and
			  int_sources < NUMBER_OF_SOURCES_PER_SYNTH 
			then
			
  		      print("Got source #" .. int_sources .. " for synth #" .. int_synths)
		      array_sample_sources_for_each_synth[int_synths][int_sources] = sample
		      array_buffer_sources_for_each_synth[int_synths][int_sources] = sample.sample_buffer
			  buffer_source = sample.sample_buffer

			  int_sources = int_sources + 1
			
		    end
			
		  end
			  
		end
		
		
		if buffer_source == nil then
		
		    print("not enough source samples in the instrument '" .. instrument.name .. "'")
		    array_bool_bad_setup[int_synths] = true
			
        end		
		
		if array_sample_destination[int_synths] == nil then
		
		  print("Sample destination insertion")
		  if instrument.name_observable:has_notifier(look_for_synths_instrument_renaming) then
		    instrument.name_observable:remove_notifier(look_for_synths_instrument_renaming)
		  end
          if instrument.samples_observable:has_notifier(look_for_synths_list_change) then
		    instrument.samples_observable:remove_notifier(look_for_synths_list_change)
		  end
		  
	      for int_sample = 1, table.getn(instrument.samples) do
	        local sample = instrument.samples[int_sample]
		
		    if sample.name_observable:has_notifier(look_for_synths_source_renaming) then
		      sample.name_observable:remove_notifier(look_for_synths_source_renaming)
		    end
		    if sample.name_observable:has_notifier(look_for_synths_destination_renaming) then
		      sample.name_observable:remove_notifier(look_for_synths_destination_renaming)
		    end		
  	      end
		  
  		  array_sample_destination[int_synths] = instrument:insert_sample_at(1)
		  array_sample_destination[int_synths].name = "Synth destination"
		  array_buffer_destination[int_synths] = array_sample_destination[int_synths].sample_buffer

		  local array_int_split = {}
		  for int_note = 1,120 do
			array_int_split[int_note] = 1
		  end
		  instrument.split_map = array_int_split
			
		  array_bool_bad_setup[int_synths] = false		
		  print("Destination created for synth #" .. int_synths)
			
		else 
		  
		  array_bool_bad_setup[int_synths] = false		
		  print("Destination found for synth #" .. int_synths)
		  
		end
		
  	    array_sample_destination[int_synths].loop_mode = array_sample_sources_for_each_synth[int_synths][1].loop_mode

        if not array_sample_destination[int_synths].name_observable:has_notifier(look_for_synths_source_renaming) then
          array_sample_destination[int_synths].name_observable:add_notifier(look_for_synths_source_renaming)
        end
		
		for int_source = 1, table.getn(array_sample_sources_for_each_synth[int_synths]) do
		
		  local sample_source = array_sample_sources_for_each_synth[int_synths][int_source]
		  if not sample_source.name_observable:has_notifier(look_for_synths_source_renaming) then
		    sample_source.name_observable:add_notifier(look_for_synths_source_renaming)
		  end
		
		end

	  end
	  
      if not instrument.samples_observable:has_notifier(look_for_synths_list_change) then
        instrument.samples_observable:add_notifier(look_for_synths_list_change)
      end
	
	end
    
  end
  
  local track_master = get_master_track()
			
  if track_master then
			
    print("Master track found")
			
	local bool_midiccdevice_found = false
	local device = nil
			
	for int_device = 1, table.getn(track_master.devices) do
			  
  	  if not bool_midiccdevice_found then
				
	    device = track_master.devices[int_device]
	    bool_midiccdevice_found = device.name == "*MIDI-CC Device"
	
	  end
				
	end
	
	if not bool_midiccdevice_found then
			  
	  print("Fake MidiCcDevice created")			
	  device = track_master:insert_device_at("Audio/Effects/    Native/*MIDI-CC Device",2)
			    
    else
			  
	  print("Fake MidiCcDevice found")
			  
    end

    local int_parameter = 0
	
    for int_synth = 1, int_synths do

      for int_source = 1, table.getn(array_buffer_sources_for_each_synth[int_synth]) do

  	    int_parameter = int_parameter + 1

	    local parameter = device.parameters[int_parameter]
		local array_parameter_info = {parameter,int_synth,int_source}
		
        parameter.value_observable:add_notifier(parameter_changed, array_parameter_info)
	    print("Added notifier for synth #" .. int_synth .. ", parameter #" .. int_source .. " on DSP slider #" .. int_parameter)
	
  	  end    
	  
    end
  
  end
	
end
			  
renoise.tool().app_new_document_observable:add_notifier(
  function()

    if renoise.song() then
	
	  look_for_synths()

      if renoise.song().instruments_observable:has_notifier(look_for_synths) then
        renoise.song().instruments_observable:remove_notifier(look_for_synths)
      end
  
      renoise.song().instruments_observable:add_notifier(look_for_synths)
  
    end
	
  end
)

renoise.tool().app_idle_observable:add_notifier(function()

  for int_synth = 1, NUMBER_OF_SYNTHS do
  
    if array_bool_mixing_requested[int_synth] then
      mix(int_synth)
	  array_bool_mixing_requested[int_synth] = false
    end
	
  end

end)

renoise.tool().app_release_document_observable:add_notifier(
  function()

    init()  

  end
)

function get_master_track()

  local track_master = nil
	
	--go to master track and add a MidiCCDevice there, if not already present
  for int_track = 1, table.getn(renoise.song().tracks) do
	
	if track_master == nil then
	  
	  local track = renoise.song().tracks[int_track]
	  if track.type == renoise.Track.TRACK_TYPE_MASTER then
		
        track_master = track		
		  
	  end
		
	end
	
  end
	
  return track_master
  
end